/*
 * Copyright (c) 2018-2019, Orbit Geospatial Technologies. All rights reserved.
 * Use is subject to license terms.
 */
define([
    "dojo/_base/declare",
    "dojo/_base/lang",
    "esri/symbols/SimpleMarkerSymbol",
    "esri/symbols/SimpleLineSymbol",
    "esri/symbols/SimpleFillSymbol",
    "esri/symbols/TextSymbol",
    "esri/graphic",
    "esri/geometry/Point",
    "esri/geometry/Polyline",
    "esri/geometry/Polygon",
    "esri/geometry/Extent",
    "esri/geometry/webMercatorUtils",
    "esri/SpatialReference",
    "esri/Color",
    "dojo/on",
    "./Orbit3DM"
], function (
    declare,
    lang,
    SimpleMarkerSymbol,
    SimpleLineSymbol,
    SimpleFillSymbol,
    TextSymbol,
    Graphic,
    Point,
    Polyline,
    Polygon,
    Extent,
    webMercatorUtils,
    SpatialReference,
    Color,
    on,
    orbitgt
) {

    // Reference some classes from the 3DMapping Viewer SDK we will be using.
    var MapObject = orbitgt.mapping3d.sdk.viewer.MapObject;
    var AList = orbitgt.mapping3d.sdk.viewer.AList;
    var Coordinate = orbitgt.mapping3d.sdk.viewer.Coordinate;
    var CoordinatePart = orbitgt.mapping3d.sdk.viewer.CoordinatePart;
    var CoordinatePath = orbitgt.mapping3d.sdk.viewer.CoordinatePath;

    // Initialize legend defaults.
    var defaultColor = new Color([0,0,255,1.0]);
    var defaultFillColor = new Color([0,0,255,0.1]);
    var defaultSymbolSize = 12.0;
    var defaultLineWidth = 1;
    var defaultFontSize = "14"; // px
    var defaultFontFamily = "Arial";
    var defaultFontColor = new Color([255,255,255,1.0]);

    var objectId = 0;

    return {

        getMapObjectForGraphic: function(graphic) {

            let geometry = graphic.geometry;
            let targetCRS = 4326;
            let targetReference = new SpatialReference({wkid:targetCRS});
            if (webMercatorUtils.canProject(geometry,targetReference))
                geometry = webMercatorUtils.project(geometry,targetReference);
            else {
                console.error("Cannot create mapobject for graphic, because we cannot project it to epsg:"+ targetCRS);
                return null;
            }

            var mapObject = new MapObject();
            mapObject.objectId = ""+(objectId++);

            if (geometry instanceof Point) {
                mapObject.spatialPoint = new Coordinate(""+targetCRS,geometry.x,geometry.y,0,false/*is3d*/);
            }
            else if (geometry instanceof Polyline) {
                mapObject.spatialLine = this.getCoordinatePathForGeometry(""+targetCRS,geometry);
            }
            else if (geometry instanceof Polygon) {
                mapObject.spatialArea = this.getCoordinatePathForGeometry(""+targetCRS,geometry);
            } else {
                console.error("Invalid geometry type: ", typeof(geometry) ,".");
                return null;
            }

            return mapObject;
        },

        getCoordinatePathForGeometry: function(targetCRS,geometry) {
            let parts = new AList(CoordinatePart.TYPE);
            let partList = geometry.paths || geometry.rings;
            partList.forEach(part => {
                let coordinates = new AList(Coordinate.TYPE);
                part.forEach(point => {
                    if (point instanceof Point)
                        coordinates.add(new Coordinate(null/*crs*/,point.x,point.y,0,false/*3d*/));
                    else if (point instanceof Array && point.length>1)
                        coordinates.add(new Coordinate(null/*crs*/,point[0],point[1],0,false/*3d*/));
                    else {
                        console.error("Cannot add point ", point);
                    }                         
                });
                if (coordinates.size()>0) {
                    let part = new CoordinatePart();
                    part.points = coordinates;
                    parts.add(part);
                }
            });
            let path = new CoordinatePath();
            path.crs = targetCRS;
            path.parts = parts;
            return path;
        },

        /**
         * Create an esri Graphic for a MapObject.
         * @param {MapObject} mapObject 
         */
        getGraphicForMapObject: function(mapObject) {
            let graphic = null;
            if (mapObject.spatialPoint) {
                graphic = new Graphic();
                graphic.geometry = new Point(mapObject.spatialPoint.x,mapObject.spatialPoint.y);
                graphic.attributes = {"mapObject":mapObject};
                graphic.symbol = this.getSymbolForPoint(mapObject);
            }
            if (mapObject.spatialLine) {
                graphic = new Graphic();
                graphic.geometry = this.getGeometryForLine(mapObject);
                graphic.attributes = {"mapObject":mapObject};
                graphic.symbol = this.getSymbolForLine(mapObject);
            }
            if (mapObject.spatialArea) {
                graphic = new Graphic();
                graphic.geometry = this.getGeometryForArea(mapObject);
                graphic.attributes = {"mapObject":mapObject};
                graphic.symbol = this.getSymbolForArea(mapObject);                    
            }
            return graphic;
        },

        /**
         * Create an esri geometry for a MapObject line.
         * @param {MapObject} mapObject 
         */
        getGeometryForLine: function(mapObject) {
            // Create path coordinates
            var parts = [];
            for (var i=0; i<mapObject.spatialLine.parts.size(); i++) {
                var coordinates = [];
                var part = mapObject.spatialLine.parts.get(i);
                for (var j=0; j<part.points.size(); j++) {
                    var point = part.points.get(j);
                    coordinates.push([point.x,point.y,point.z]);
                }
                parts.push(coordinates);
            }
            // Create polyline
            return new Polyline(parts.length==1 ? parts[0] : parts);
        },

        /**
         * Create an esri geometry for a MapObject area.
         * @param {MapObject} mapObject 
         */
        getGeometryForArea: function(mapObject) {
            // Create parts of lists of coordinates.
            var parts = [];
            for (var i=0; i<mapObject.spatialArea.parts.size(); i++) {
                var coordinates = [];
                var part = mapObject.spatialArea.parts.get(i);
                for (var j=0; j<part.points.size(); j++) {
                    var point = part.points.get(j);
                    coordinates.push([point.x,point.y,point.z]);
                }
                parts.push(coordinates);
            }
            // Create polygon
            return new Polygon(parts.length==1 ? parts[0] : parts);
        },

        /**
         * Create an esri symbol for rendering a MapObject label.
         * @param {MapObject} mapObject 
         */
        getSymbolForLabel: function(mapObject) {
            var color = defaultFontColor;
            var fontSize = defaultFontSize;
            var fontFamily = defaultFontFamily;

            if (mapObject.labelFontSize)
                fontSize = mapObject.labelFontSize;

            if (mapObject.labelFontFamily)
                fontFamily = mapObject.labelFontFamily;

            if (mapObject.labelColor) {
                var a = mapObject.labelColor.alpha/255;
                var r = mapObject.labelColor.red;
                var g = mapObject.labelColor.green;
                var b = mapObject.labelColor.blue;
                color = new Color([r,g,b,a]);
            }

            var font = new esri.symbol.Font(fontSize+"px",
            esri.symbol.Font.STYLE_NORMAL,
            esri.symbol.Font.VARIANT_NORMAL,
            esri.symbol.Font.WEIGHT_BOLD, fontFamily);

            var symbol = new TextSymbol(mapObject.label,font,color);
            symbol.horizontalAlignment = TextSymbol.ALIGN_MIDDLE;
            symbol.verticalAlignment = TextSymbol.ALIGN_MIDDLE;
            if (mapObject.labelPosition) 
                symbol.setOffset(mapObject.labelPosition.x,-mapObject.labelPosition.y);
            return symbol;
        },

        /**
         * Create an esri symbol for rendering a MapObject point.
         * @param {MapObject} mapObject 
         */
        getSymbolForPoint: function(mapObject) {

            var fillColor = defaultFillColor;
            var symbolSize = mapObject.symbolSize || defaultSymbolSize;

            if (mapObject.fillColor) {
                var a = mapObject.fillColor.alpha/255;
                var r = mapObject.fillColor.red;
                var g = mapObject.fillColor.green;
                var b = mapObject.fillColor.blue;
                fillColor = new Color([r,g,b,a]);
            }

            if (mapObject.symbolPath) {

                var parts = this.getSymbolPath(mapObject);
                var maxDistance = parts[0];
                var symbolPath = parts[1];
    
                var symbol = new SimpleMarkerSymbol();
                symbol.setSize(symbolSize);
                symbol.setPath(symbolPath);
                symbol.setColor(fillColor)
                symbol.setOutline(this.getSymbolForLine(mapObject));
                return symbol;
                
            } else {

                // The default symbol.
                var symbolName = SimpleMarkerSymbol.STYLE_CIRCLE;
                return new SimpleMarkerSymbol(symbolName,symbolSize,this.getSymbolForLine(mapObject),fillColor);

            } 
        },

        /**
         * Create an esri symbol for rendering a MapObject line.
         * @param {MapObject} mapObject 
         */
        getSymbolForLine: function(mapObject) {

            var color = defaultColor;
            var lineWidth = mapObject.lineWidth || defaultLineWidth;

            if (mapObject.color) {
                var a = mapObject.color.alpha/255;
                var r = mapObject.color.red;
                var g = mapObject.color.green;
                var b = mapObject.color.blue;
                color = new Color([r,g,b,a]);
            }

            return new SimpleLineSymbol(
                SimpleLineSymbol.STYLE_SOLID,
                color,
                lineWidth
            );
        },

        /**
         * Create an esri symbol for rendering a MapObject area.
         * @param {MapObject} mapObject 
         */
        getSymbolForArea: function(mapObject) {
            var fillColor = defaultFillColor;

            if (mapObject.fillColor) {
                var a = mapObject.fillColor.alpha/255;
                var r = mapObject.fillColor.red;
                var g = mapObject.fillColor.green;
                var b = mapObject.fillColor.blue;
                fillColor = new Color([r,g,b,a]);
            }

            return new SimpleFillSymbol(
                SimpleFillSymbol.STYLE_SOLID,
                this.getSymbolForLine(mapObject),
                fillColor
            );
        },

        /**
         * Create an SVG path from a MapObject symbol path.
         * @param {MapObject} mapObject 
         */
        getSymbolPath: function(mapObject) {

            if (!mapObject.symbolPath) return null;

            var maxX = 0;
            var maxY = 0;

            // create path
            var path = "";
            for (var i=0; i<mapObject.symbolPath.parts.size(); i++) {
                var part = mapObject.symbolPath.parts.get(i);
                for (var j=0; j<part.points.size(); j++) {
                    var point = part.points.get(j);

                    maxX = Math.max(Math.abs(point.x),maxX);
                    maxY = Math.max(Math.abs(point.y),maxY);

                    var type = j==0 ? "M" : "L";
                    path += type+point.x+","+point.y;
                }
            }

            // create contour
            var maxDistance = Math.max(maxX,maxY);
            var list = [-maxDistance,maxDistance,maxDistance,maxDistance,maxDistance,-maxDistance,-maxDistance,-maxDistance];
            var contour = "M"+list[0]+","+list[1]+"M"+list[2]+","+list[3]+"M"+list[4]+","+list[5]+"M"+list[6]+","+list[7];
            return [maxDistance,contour+path];
        }
    };
});
