/*
 * Copyright (c) 2018-2019, Orbit Geospatial Technologies. All rights reserved.
 * Use is subject to license terms.
 */
define([
    "dojo/ready",
    "dojo/_base/declare",
    "dojo/_base/lang",
    "esri/dijit/Popup",
    "esri/layers/GraphicsLayer",
    "esri/layers/LabelLayer",
    "esri/graphic",
    "esri/geometry/Point",
    "esri/geometry/Extent",
    "esri/SpatialReference",
    "dojo/on",
    "esri/geometry/scaleUtils",
    "esri/geometry/webMercatorUtils",
    "./MapObjectUtils",
    "jimu/BaseWidget",
    "./Orbit3DM"
], function (
    ready,
    declare,
    lang,
    Popup,
    GraphicsLayer,
    LabelLayer,
    Graphic,
    Point,
    Extent,
    SpatialReference,
    on,
    scaleUtils,
    webMercatorUtils,
    mapObjectUtils,
    BaseWidget,
    orbitgt
) {

    // Reference some classes from the 3DMapping Viewer SDK we will be using.
    var SDKViewer = orbitgt.mapping3d.sdk.viewer.SDKViewer;
    var Bounds = orbitgt.mapping3d.sdk.viewer.Bounds;
    var Coordinate = orbitgt.mapping3d.sdk.viewer.Coordinate;
    var MapObject = orbitgt.mapping3d.sdk.viewer.MapObject;
    var AList = orbitgt.mapping3d.sdk.viewer.AList;
    var AMap = orbitgt.mapping3d.sdk.viewer.AMap;
    var Constants = orbitgt.mapping3d.sdk.viewer.Constants;

    // To create a widget, you need to derive from BaseWidget.
    return declare([BaseWidget], {

        postCreate: function () {
            this.inherited(arguments);
        },

        /**
         * Called on widget startup. 
         */
        startup: function() {
            // Initialize a map of MapObject layers.
            this.mapObjectLayers = new Map();
        },

        /**
         * Called when the widget is opened. 
         */
        onOpen: function() {
            // The viewer is not created yet?
            if (this.viewer==null) {

                // Prepare startup options.
                var options = new AMap();
                options.set(Constants.STARTUP_APP_STATE,window.localStorage.getItem("orbit_mapping3d_state"));
                options.set(Constants.STARTUP_CRS,"4326");
                options.set(Constants.STARTUP_THROTTLE_ANIMATION_SIGNALS,true);
                // Choose an application name that best suits your plugin
                var applicationName = "Orbit 3D Mapping Web AppBuilder Widget";
                // Create viewer instance
                this.viewer = new SDKViewer(applicationName,this.mmPanel,options);
                window.viewer = this.viewer;
                window.orbitgt = orbitgt;
                // Wait for the isReady promise...
                this.viewer.isReady.then((viewer) => {
                    
                    console.log("Viewer is ready!",viewer);
    
                    // Update size.
                    this.updateSize();
    
                    // Below we create some of the layers we will need, ahead of time.
                    // We do this now, so we can be sure they are added to the Map in the correct sequence.
                    this.findLayer(MapObject.TYPE_FOOTPRINT);
                    this.findLayer(MapObject.TYPE_MEASUREMENTS);
                    this.findLayer(MapObject.TYPE_GLOBAL);
                    this.findLayer(MapObject.TYPE_FOV);
    
                    // Listen to map-interaction signals 
                    this.viewer.onZoomToExtent.add(lang.hitch(this,this.handleZoomToExtent));
                    this.viewer.onAppStateChanged.add(lang.hitch(this,this.saveState));
                    this.viewer.onPickPositionRequested.add(lang.hitch(this,this.handlePickPositionRequested));
                    this.viewer.onPickPositionCancelled.add(lang.hitch(this,this.handlePickPositionCancelled));
                    this.viewer.onWorkspaceOpened.add(lang.hitch(this,this.handleWorkspaceOpened));
    
                    // Listen to the esri map signals
                    this.own(on(this.map,"extent-change",lang.hitch(this,this.updateFootprints)));
                    this.own(on(this.map,"click",lang.hitch(this,this.onMapClick)));
    
                    // Listen to resource signals
                    this.viewer.onFootprintsChanged.add(lang.hitch(this,this.updateFootprints));
                    this.viewer.onGlobalsChanged.add(lang.hitch(this,this.updateGlobals));
                    this.viewer.onMeasureStateChanged.add(lang.hitch(this,this.updateMeasurements));
                    this.viewer.onFieldOfViewsChanged.add(lang.hitch(this,this.updateViews));

                    if (this.map.infoWindow instanceof Popup) {
                        this.own(on(this.map.infoWindow,"selection-change",lang.hitch(this,this.handleSelectionChange)));
                    } else {
                        console.log("infoWindow not a Popup instance, unable to track selection.", this.map.infoWindow);
                    }

                    this.viewer.onOverlayObjectChanged.add(lang.hitch(this,this.handleOverlayObjectChanged));
                })
            }
        },

        handleOverlayObjectChanged: function(object) {
            console.log("Overlay object changed: ", object);
        },

        handleWorkspaceOpened: function() {
            console.log("Workspace is open.");
            // this.viewer.doAction("startMeasure:point")
        },

        handleSelectionChange: function() {
            let graphic = this.map.infoWindow.getSelectedFeature();
            let mapObject = null;
            if (graphic) {
                mapObject = mapObjectUtils.getMapObjectForGraphic(graphic);
                mapObject.objectId = "selectedObject";
            }

            this.viewer.clearOverlayObjects();         

            let list = new AList(MapObject.TYPE);
            if (mapObject) list.add(mapObject);
            this.viewer.addOverlayObjects(list);
        },

        /**
         * Called on widget ready.
         */
        onReady: function () {
            this.ready = true;
        },

        /** 
         * Called on widget resize.
         */
        resize: function() {
            this.updateSize();
        },

        /** 
         * This function returns a layer for a specific MapObject type.
         * @param mapObjectType the MapObject type.
         * @param layerType the esri layer type (null of GraphicsLayer, "label" for LabelLayer).
         * @return returns an esri Layer instance, already added to the map.
         */
        findLayer: function(mapObjectType,layerType) {
            // The layer does not exist?
            if (!this.mapObjectLayers.has(mapObjectType)) {
                // Create a new layer
                var layer = null;
                if (layerType=="label") layer = new LabelLayer();
                else layer = new GraphicsLayer();
                // Save a reference
                this.mapObjectLayers.set(mapObjectType,layer);
                // Add to map
                this.map.addLayer(layer);
            }
            // Return the layer
            return this.mapObjectLayers.get(mapObjectType);
        },

        updateGlobals: function() {
            // Find the map layers.
            var layer = this.findLayer(MapObject.TYPE_GLOBAL);
            // Clear objects
            layer.clear();
            // Add objects 
            let mapObjects = this.viewer.getGlobals();
            this.addMapObjectsToLayer(layer,null,mapObjects);
            // Repaint
            layer.redraw();
        },

        updateMeasurements: function() {
            // Find the map layers.
            var layer = this.findLayer(MapObject.TYPE_MEASUREMENT);
            // Clear objects
            layer.clear();
            // Add objects 
            let measureState = this.viewer.getMeasureState();
            this.addMapObjectsToLayer(layer,null,measureState.mapObjects);
            // Repaint
            layer.redraw();        
        },

        /**
         * Update the resource footprints.
         */
        updateFootprints: function() {

            // First call? Initialize some flags.
            if (this.isUpdatingFootprints==undefined) {
                this.isUpdatingFootprints = false;
                this.isUpdatingFootprintsPending = false;
            }

            // Updating footprints?
            if (this.isUpdatingFootprints==true) {
                // Schedule a pending update and bail out.
                this.isUpdatingFootprintsPending = true;
                return;
            }

            // Prepare a closure that calls this function again if isUpdatingFootprintsPending is true.
            var handleFinished = lang.hitch(this,function() {
                this.isUpdatingFootprints = false;
                if (this.isUpdatingFootprintsPending==true) {
                    this.isUpdatingFootprintsPending = false;
                    this.updateFootprints();
                }
            });

            // We are updating.
            this.isUpdatingFootprints = true;
            // Find the map layers.
            var layer = this.findLayer(MapObject.TYPE_FOOTPRINT);
            var labelLayer = this.findLayer(MapObject.TYPE_FOOTPRINT+"-label","label"/*type*/);
            // Get map extent
            var extent = this.map.geographicExtent;
            var bounds = Bounds.fromXY("4326",extent.xmin,extent.ymin,extent.xmax,extent.ymax);
            // Load footprint data
//            console.log("scale: "+ scaleUtils.getScale(this.map));
            this.viewer.getFootprints2D(bounds,scaleUtils.getScale(this.map))
                .then(mapObjects => {
                    // Clear objects
                    layer.clear();
                    labelLayer.clear();
                    // Add objects 
                    this.addMapObjectsToLayer(layer,labelLayer,mapObjects);
                    // Repaint
                    layer.redraw();
                    labelLayer.redraw();
                })
                .then(result => {
                    handleFinished();
                }, reason => {
                    console.error("Error calling getFootprints2D(",bounds,")",reason);
                    handleFinished();
                });
        },

        /**
         * Update the viewer field-of-views.
         */
        updateViews: function() {
            // Find the map layers
            var layer = this.findLayer(MapObject.TYPE_FOV);
            var labelLayer = this.findLayer(MapObject.TYPE_FOV+"-labels");
            // Clear objects
            layer.clear();
            labelLayer.clear();
            // Get field-of-view objects
            var mapObjects = this.viewer.getFieldOfViews2D();
            // Add objects
            this.addMapObjectsToLayer(layer,labelLayer,mapObjects);
            // Repaint
            layer.redraw();
            labelLayer.redraw();
        },

        /** 
         * Generic function that adds a list of MapObject instances to a map layer.
         * @param {GraphicLayer} layer a layer for Graphic objects.
         * @param {LabelLayer} labelLayer a layer for labels.
         * @param {AList} mapObjects the list of map objects.
         */
        addMapObjectsToLayer: function(layer,labelLayer,mapObjects) {

            // Add objects
            for (var i=0; i<mapObjects.size(); i++)
            {
                var mapObject = mapObjects.get(i);
                var graphic = mapObjectUtils.getGraphicForMapObject(mapObject);
                var labelGraphic = null;

                if (graphic) {
                    // The MapObject has a label and we have a layer for labels?
                    if (mapObject.label && labelLayer) {
                        // Create a label graphic. 
                        labelGraphic = new Graphic();
                        // Attach the label to the graphic's position (in case of a point) 
                        // or else to the graphic extent's center.
                        if (graphic.geometry instanceof Point) labelGraphic.geometry = graphic.geometry;
                        else labelGraphic.geometry = graphic.geometry.getExtent().getCenter();
                        // Set the label symbol.
                        labelGraphic.symbol = mapObjectUtils.getSymbolForLabel(mapObject);
                    }
                }                

                // Add objects to layers.
                if (graphic) layer.add(graphic);
                if (labelGraphic) labelLayer.add(labelGraphic);
            }
        },

        /**
         * Handle a "zoom to extent" signal from the viewer.
         */
        handleZoomToExtent: function(extent) {
            // Create esri Extent
            var extent = new Extent(
                extent.minX,
                extent.minY,
                extent.maxX,
                extent.maxY,
                new SpatialReference({wkid:4326})
            );
            if (webMercatorUtils.canProject(extent,this.map)) {
                // Project to map crs
                extent = webMercatorUtils.project(extent,this.map);
                // Set extent
                this.map.setExtent(extent);
            }
        },

        /**
         * Handle a "pick position on the map" signal from the viewer.
         */
        handlePickPositionRequested: function() {
            this.map.setMapCursor("crosshair");
            this.waitingForPickPosition = true;
        },

        /**
         * Handle a "pick position on the map" signal from the viewer.
         */
        handlePickPositionCancelled: function() {
            this.map.setMapCursor("default");
            this.waitingForPickPosition = false;
        },

        /** 
         * Handle a "application state changed" signal from the viewer.
         */
        saveState: function(state) {
            // save it.
            window.localStorage.setItem("orbit_mapping3d_state",state);
        },

        /**
         * Update the viewer to the available size of the widget.
         */
        updateSize: function() {
            if (this.viewer)
                this.viewer.setSize(this.mmPanel.clientWidth,this.mmPanel.clientHeight);
        },

        /**
         * Handle an esri map click.
         */
        onMapClick(event) {

            var mapPoint = event.mapPoint;
            var wgsSpatialRef = new SpatialReference({wkid:4326});
            var coordinate;

            if (webMercatorUtils.canProject(mapPoint.spatialReference,wgsSpatialRef)) {
                let mapPoint2 = webMercatorUtils.project(mapPoint,wgsSpatialRef);
                coordinate = Coordinate.fromXY("4326",mapPoint2.x,mapPoint2.y);
            }

            if (this.waitingForPickPosition==true) {

                // Pass the position on to the viewer.
                if (coordinate)
                    this.viewer.setPickPosition(coordinate);
                // Clear cursor and flag.
                this.map.setMapCursor("auto");
                this.waitingForPickPosition = false;

            } 
        }, 
    });
});
